use clap::{Args, ValueEnum};
use std::{env, path::PathBuf, process::Command, str};
use tool_helper::Error;

#[derive(Args)]
pub struct Arguments {
    #[clap(value_enum, value_parser, default_value_t=Quality::High)]
    quality: Quality,
    #[clap(value_enum, value_parser, default_value_t=Sample::Stereo)]
    sample: Sample
}

#[derive(Copy, Clone, ValueEnum)]
pub enum Quality {
    Low,
    High
}

impl Quality {
    fn get_frequency(&self) -> u64 {
        match self {
            Self::Low  => 18900,
            Self::High => 37800,
        }
    }
}

#[derive(Copy, Clone, ValueEnum)]
pub enum Sample {
    Mono,
    Stereo
}

impl Sample {
    fn get_channel(&self) -> u64 {
        match self {
            Self::Mono   => 1,
            Self::Stereo => 2,
        }
    }
}

pub fn convert(args: Arguments, input: PathBuf, output: PathBuf) -> Result<(), Error> {
    let quality   = args.quality;
    let sample    = args.sample;
    let tool_path = {
        let mut my_path = env::current_exe()?;

        my_path.pop();
        my_path.push("extern");
        my_path.push("psxavenc");

        my_path
    };

    if !tool_path.exists() {
        return Err(Error::from_str("Could not locate psxavenc. Make sure it is installed under \"<Jaby Engine>/bin/extern\""));
    }

    let result = Command::new(tool_path).args([
        "-t", "xacd", 
        "-f", quality.get_frequency().to_string().as_ref(),
        "-b", "4",
        "-c", sample.get_channel().to_string().as_ref(),
        "-F", "0",
        "-C", "0",
        input.to_string_lossy().as_ref(), output.to_string_lossy().as_ref()
    ]) .output()?;

    let stderr = tool_helper::vec_helper::to_string(result.stderr)?;
    let stdout = tool_helper::vec_helper::to_string(result.stdout)?;

    if !result.status.success() {
        return Err(Error::from_text(format!("psxavenc returned: {}. {}", result.status, stderr)));
    }

    if !stdout.is_empty() {
        println!("{}", stdout);
    }
    Ok(())
}